importPackage(com.change_vision.jude.api.inf.editor);
importPackage(com.change_vision.jude.api.inf.presentation);
importPackage(com.change_vision.jude.api.inf.model);
importPackage(com.change_vision.jude.api.inf.project);
importPackage(java.util);

var ID_PREFIXES = {'goal': 'G',
                    'strategy': 'S',
                    'solution': 'Sn',
                    'context': 'C',
                    'justification': 'J',
                    'assumption': 'A',
                    'module': 'M'};
var INITIAL_COUNTER_NUMBER = 1;  //Begin from G0 or G1
var counterForId;
var TEMP_ID_PREFIX = '#';  //Prefix to avoid conflicted ID
var isTemporaryModification = true;
var nodesHaveReset = new HashSet(); //To avoid updating the id many times
var modelMap = new HashMap(); //To search model by name <fullNameString, INamedElement>
var originalIdMap = new HashMap(); //To remember the original ID to show the result
var ID_LABEL_SEPARATOR = ' ' //To keep the last part of ID (ex. [G123 Label]-->[G1 Label]

run();

function run() {
    var nodes = getGsnNodesOnDiagram();
    if (nodes.length === 0) return;
    
    prepareToResetId(nodes);
    
    var roots = getTreeRoots(nodes);
    resetId(roots);
    
    showResult();
}

function getGsnNodesOnDiagram() {
    var presentations = getAllPresentationsInDiagramEditor();
    if (presentations.length === 0) {
        return [];
    }
    var nodes = presentations.filter(function(p) {
        return isGsnNode(p.model);
    });
    return nodes;
}

function prepareToResetId(nodes) {
    clearCounter();
    updateModelMap();
    rememberOriginalId(nodes);
}

function resetId(roots) {
    TransactionManager.beginTransaction();
    
    //Reset ID with temporary prefix to avoid conflict at first
    isTemporaryModification = true;
    roots.forEach(function(root) {
        setNewId(root);
        resetChildrenId(root, new HashSet());
    });
    
    clearCounter();
    updateModelMap();

    //Reset ID without temporary prefix
    isTemporaryModification = false;
    roots.forEach(function(root) {
        setNewId(root);
        resetChildrenId(root, new HashSet());
    });

    TransactionManager.endTransaction();
}

function showResult() {
    originalIdMap.keySet().toArray().sort(function(node1, node2) {
        if (originalIdMap.get(node1) < originalIdMap.get(node2)) {
            return -1;
        } else if (originalIdMap.get(node1) > originalIdMap.get(node2)) {
            return 1;
        } else {
            return 0;
        }
    }).forEach(function(node) {
        var originalId = originalIdMap.get(node);
        var id = node.model.name;
        if (originalId != id) {
            println('Modified: ' + originalId + ' ---> ' + id);
        }
    });
    println('Finished');
}

function rememberOriginalId(nodes) {
    nodes.forEach(function(node) {
        originalIdMap.put(node, node.model.name);
    });
}

function clearCounter() {
    nodesHaveReset.clear();
    counterForId = {'goal': INITIAL_COUNTER_NUMBER,
                    'strategy': INITIAL_COUNTER_NUMBER,
                    'solution': INITIAL_COUNTER_NUMBER,
                    'context': INITIAL_COUNTER_NUMBER,
                    'justification': INITIAL_COUNTER_NUMBER,
                    'assumption': INITIAL_COUNTER_NUMBER,
                    'module': INITIAL_COUNTER_NUMBER};
}

function updateModelMap() {
    modelMap.clear();
    var models = astah.findElements(new ModelFinder() {
       isTarget: function(element) {
           if (element instanceof IClass || element instanceof IPackage) {
               return true;
           }
           return false;
       }
    });
    
    models.forEach(function(m) {
        modelMap.put(m.getFullName('::'), m);
    });
}

function resetChildrenId(node, traversedNodes) {
    traversedNodes.add(node);
    var contexts = getContexts(node);
    contexts.forEach(function(context) {
        setNewId(context);
    });
    
    var children = getChildren(node);
    children.forEach(function(child) {
        setNewId(child);
    });
    children.forEach(function(child) {
        if (traversedNodes.contains(child)) return;
        resetChildrenId(child, traversedNodes);
    });
}

function setNewId(node) {
    if (nodesHaveReset.contains(node)) return;
    var newId = getNewIdWithoutNameConflict(node);
    //println(node.getModel().getName() + ' --> ' + newId);
    node.getModel().setName(newId);
    nodesHaveReset.add(node);
}

function getNewIdWithoutNameConflict(node) {
    var newId = getNewId(node);
    while (hasNameConflict(node, newId)) {
        println('Skipped [' + newId + '] to avoid conflict');
        newId = getNewId(node);
    }
    return newId;
}

function hasNameConflict(node, newId) {
    var newFullName = newId;
    var fullNamespace = node.getModel().getFullNamespace('::');
    if (fullNamespace.length() > 0) {
        newFullName = fullNamespace + '::' + newId;
    }
    if (modelMap.get(newFullName) != null) {
        return true;
    }
    return false;
}

function getNewId(node) {
    var newId = '';
    var model = node.model;
    if (isGoal(model)) {
        newId = getNewIdString('goal');
    } else if (isStrategy(model)) {
        newId = getNewIdString('strategy');
    } else if (isSolution(model)) {
        newId = getNewIdString('solution');
    } else if (isModule(model)) {
        newId = getNewIdString('module');
    } else if (isContext(model)) {
        newId = getNewIdString('context');
    } else if (isAssumption(model)) {
        newId = getNewIdString('assumption');
    } else if (isJustification(model)) {
        newId = getNewIdString('justification');
    } else {
        println('Impossible!');
        return;
    }
    //To keep label
    var originalLabel = getOriginalLabel(model);
    if (originalLabel != null) {
        newId = newId + ID_LABEL_SEPARATOR + getOriginalLabel(model);
    }

    if (isTemporaryModification) {
        newId = TEMP_ID_PREFIX + newId;
    }
    return newId;
}

function getNewIdString(nodeKind) {
    var counter = counterForId[nodeKind];
    counterForId[nodeKind] = counter + 1;
    return ID_PREFIXES[nodeKind] + counter;
}

function getOriginalLabel(model) {
    var originalName = model.name;
    var separatorIndex = originalName.indexOf(ID_LABEL_SEPARATOR);
    if (separatorIndex >= 0 && separatorIndex + ID_LABEL_SEPARATOR.length < originalName.length()) {
        
        return originalName.substring(separatorIndex + ID_LABEL_SEPARATOR.length);
    }
    return null;
}

function getAllPresentationsInDiagramEditor() {
    var diagramViewManager = astah.viewManager.diagramViewManager;
    var diagram = diagramViewManager.currentDiagram;
    if (diagram == null) {
        println('Please open a diagram.');
        return [];
    }
    return diagram.presentations;
}

function getTreeRoots(presentations) {
    var roots = new HashSet();
    presentations.forEach(function(p) {
        roots.add(getTreeRoot(p, new HashSet()));
    });
    var rootsArray = roots.toArray();
    //sort by the location
    rootsArray.sort(function(a, b) {
        return a.location.x - b.location.x;
    });
    return rootsArray;
}

function getTreeRoot(node, traversedNodes) {
    traversedNodes.add(node);
    var parent = getParent(node);
    if (parent == null) {
        return node;
    } else {
        if (traversedNodes.contains(parent)) return node;
        return getTreeRoot(parent, traversedNodes);
    }
}

function getChildren(node) {
    var links = node.links;
    var children = [];
    links.forEach(function(link) {
        if (isSupportedBy(link.model) && link.target == node) {
            children.push(link.source);
        }
    });
    
    //sort by the location
    children.sort(function(a, b) {
        return a.location.x - b.location.x;
    });
    return children;
}

/* Return Context, Assumption and Justification */
function getContexts(node) {
    var links = node.links;
    var contexts = [];
    links.forEach(function(link) {
        if (isInContextOf(link.model) && link.target == node) {
            contexts.push(link.source);
        }
    });
    
    //sort by the location
    contexts.sort(function(a, b) {
        return a.location.y - b.location.y;
    });
    return contexts;
}

function getParent(node) {
    var links = node.links;
    //check supportedBy first
    for (var i = 0; i < links.length; i++) {
        var link = links[i];
        if (isSupportedBy(link.model) && link.source == node) {
            return link.target;
        }
    }
    //check inContextOf next
    for (var i = 0; i < links.length; i++) {
        var link = links[i];
        if (isInContextOf(link.model) && link.source == node) {
            return link.target;
        }
    }
    return null;
}

function isSupportedBy(model) {
    return getStereotype(model) == 'supportedBy';
}

function isInContextOf(model) {
    return getStereotype(model) == 'inContextOf';
}

function isGoal(model) {
    return getStereotype(model) == 'goal';
}

function isStrategy(model) {
    return getStereotype(model) == 'strategy';
}

function isSolution(model) {
    return getStereotype(model) == 'solution';
}

function isModule(model) {
    return getStereotype(model) == 'module';
}

function isContext(model) {
    return getStereotype(model) == 'context';
}

function isAssumption(model) {
    return getStereotype(model) == 'assumption';
}

function isJustification(model) {
    return getStereotype(model) == 'justification';
}
 
function isGsnNode(model) {
    if (isGoal(model) || isStrategy(model) || isSolution(model)
        || isContext(model) || isJustification(model)
        || isAssumption(model) || isModule(model)) {
        return true;
    }
    return false;
}

function getStereotype(model) {
    if (model == null) return '';
    var stereotypes = model.getStereotypes();
    if (stereotypes != null && stereotypes.length > 0) {
        return stereotypes[0];
    }
    return '';
}